/*
#define RCS_ID $Id: pb.c,v 1.4 1997/06/21 07:38:06 oz7sat Exp $
 *   Copyright 1992, 1993, 1994 John Melton (G0ORX/N6LYT)
 *              All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   This program is modified by Peter Scott oz2aba/oz7sat @ amsat.org
 *   The program is extended with:
 *	No X11 support
 *	Uses ncurses to handle display control
 *	Automatic update of directory
 *	Database of files to be downloaded
 *	Extended control of pb by UDP packets
 *	Automatic selection of files to download based on a rule file (not yet)
 *   
 */

/*
	pb.c

	Pacsat Broadcast Protocal Receiver

	This software was written to provide a receive only capability
	for collecting files and directories from satellites running
	the Pacsat Protocols running on the Linux Operating System.

	The TNC must be setup for KISS.
*/

/*	Changes (OZ6BL) :

	V 1.1

	-	Changed strategy for sending fill requests to the satellite:
		When a frame addressed to PBLIST is received:
		IF I'm in the queue THEN
		   IF my request is a file request THEN
			do nothing
		   ELSE
			send a file request (if any)
		ELSE
		   send any request (file or directory)

	-	Changed name of kiss log file to YYMMDDHH.kss
		(e.g. 96123014.kss)

	-	Reverted back to use of sockets

*/

#define VERSION_STRING "(version 1.1 by oz2aba and oz6bl)"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <linux/if_ether.h>
#include <linux/ax25.h>

#include <fcntl.h>
/* #include <termios.h> */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <strings.h>

#include <curses.h>

#include "ftl0.h" 
#include "pb.h"
#include "header.h"
#include "request.h"
#include "crc.h"

#include "axconfig.h"
#include "axutils.h"

#define REQUEST_NONE 0
#define REQUEST_FILE 1
#define REQUEST_ANY 2

#define	FILL_NONE	0
#define	FILL_UNIQUE	1
#define	FILL_OVERLAP	2

FILE *fpkiss = NULL ;

char writebuf[1000];

long id;		/* ID number of the file in pursuit */

/* Program parameters */
char Satellite   [100];
char SatName     [100];
char MyCall      [100];
char MaxDays     [100];
char UDPPort     [100];
char KissLog     [100];

unsigned int UdpPort     = 5100;
int	     kisslog     = 0;

unsigned long int curByteRead   = 0;
unsigned long int curByteWrite  = 0;
unsigned long int lastByteRead  = 0;
unsigned long int lastByteWrite = 0;

int s_input;
int s_file;
int s_directory;
struct sockaddr_ax25	src;
struct sockaddr_ax25	dest;

struct sockaddr_in request_addr;
int s_request;		/* Handle for receiving internal requests */

#define	TIMEOUT		2
#define TIMEOUTFLUSH	30

unsigned char	buf [MAXBUFFER];
int		bufSize;


OPENFILES of[MAXFILES + 1];	/* last entry is for the directory file */
int nextof = 0;

unsigned long maxDays = 5;
int stopFlag = 0;		/* Set to nonzero to terminate program */
int traceFlag = 0;		/* Controls frame trace. 0:None, 1:Head, 2:All */
int directoryOld = 1;		/* Marks the directory as in unknown state */

int bytes     = 0;		/* Trafic counters */
int fileBytes = 0;
int dirBytes  = 0;
int tlmBytes  = 0;
int crcErrors = 0;

time_t start_time = 0;		/* Used for Bytes/Second/Pass calculations */
time_t end_time   = 0;

/* Screen control pointers */
WINDOW *scr_download, *scr_files, *scr_log, *scr_status;

/* Download files control structures */
unsigned long download_list [MAXDOWNLOAD];
unsigned long download_active = 0xFFFFFFFF;

LOADLIST	*autoload_list = NULL;


void FillDirectory(void);
void FillFile(int *);
void CancelFill(void);

/*
 *	Limit the contents of text to printable chars
 */
void notprintable (char *text)
{
	text--;
	while (*(++text) != '\0') {
		if (*text >= ' ' && *text <= '}') continue;
		if (*text == '\t') continue;
		if (*text == '\r') *text = '\n';
		if (*text == '\n') continue;
		*text = '.';
	}
}


/*
 *	Write a text line on one of the log screen
 */
void writelog (int scrno, char *text)
{
	/* Remove the tail if it is a \n */
	while (text [strlen (text)-1] == '\n' || text [strlen (text)-1] == ' ')
		text [strlen (text) - 1] = '\0';

	if (scrno == 1) {
		wprintw (scr_log, "\n%s", text);
		wrefresh (scr_log);
	} else {
		wprintw (scr_files, "\n%s", text);
		wrefresh (scr_files);
	}

	refresh ();
}

/*
 *	Write a text line on the download screen
 */
void writedownload (int lin, char *text)
{
	mvwprintw (scr_download, lin, 0, "%s", text);
	wclrtoeol (scr_download);
	wrefresh (scr_download);
}

/*
 *	Write a string on the status line
 */
void writestatus (int col, char *text)
{
	mvwprintw (scr_status, 0, col, "%s", text);
	wrefresh (scr_status);
}

/*
 *	Write an ERROR message
 *
 */
void MessageBox(char *text)
{
	char	s [100];
	sprintf (s, "ERROR: %s", text);
	writelog (1, "");
	writelog (1, s);
	writelog (1, "");
}


/*
 *	Convert a call from the shifted ascii form used in an
 *	AX.25 packet.
 */
int ConvertCall(char *c, char *call)
{
	char *ep = c + 6;
	int ct = 0;

	while (ct < 6)
	{
		if (((*c >> 1) & 127) == ' ') break;

		*call = (*c >> 1) & 127;
		call++;
		ct++;
		c++;
	}
	
	if ((*ep & 0x1E) != 0)
	{	
		*call = '-';
		call++;
		call += sprintf(call, "%d", (int)(((*ep) >> 1) & 0x0F));
	}

	*call = '\0';
	
	if (*ep & 1) return 0;

	return 1;
}

/*
 *	Convert a call to the shifted ascii form used in an
 *	AX.25 packet.
 */
int MakeAddress(char *name, struct sockaddr_ax25 *sax)
{
	int ct = 0;
	int ssid = 0;
	char *p = name;

	while (ct < 6)
	{
		if (*p == '-' || *p == '\0') break;

		if (islower(*p)) *p = toupper(*p);

		if (!isalnum(*p))
		{
			printf("Invalid symbol in callsign %c\n", *p);
			return -1;
		}
		
		sax->sax25_call.ax25_call[ct] = (*p << 1);
		p++;
		ct++;
	}

	while (ct < 6)
	{
		sax->sax25_call.ax25_call[ct] = ' ' << 1;
		ct++;
	}

	if (*p == '-')
	{
		p++;

		if (sscanf(p, "%d", &ssid) != 1 || ssid < 0 || ssid > 15)
		{
			printf("Invalid SSID - %s - %s\n", name, p);
			return -1;
		}
	}

	sax->sax25_call.ax25_call[6]=((ssid + '0') << 1) & 0x1E;
        sax->sax25_family = AF_AX25 ;

	return 0;
}

/*
 *	Load/Reload the download.req file
 */
void CheckDownloadFile (void)
{
	FILE	*fp;
	int	x;

	fp = fopen ("download.req", "r");
	if (fp == NULL) {		/* No download file. Create it */
		memset (download_list, 0xFF, sizeof (unsigned long) * MAXDOWNLOAD);
		fp = fopen ("download.req", "w");
		fwrite (download_list, sizeof (unsigned long), MAXDOWNLOAD, fp);
		fclose (fp);
		return;
	}

	fread (download_list, sizeof (unsigned long), MAXDOWNLOAD, fp);
	x = ferror (fp);
	fclose (fp);
	if (x) return;			/* Error reading. Ignore it */

	/* If download_active is -1 find the oldest file to download */
	if (download_active != 0xFFFFFFFF) return;

	for (x = 0, download_active = 0xFFFFFFFF; x < MAXDOWNLOAD; x++) {
		if (download_list [x] < download_active)
			download_active = download_list [x];
	}
}

/*
 *	Remove a file from the download_list and file
 */
void RemoveDownloadFile (unsigned long file_id)
{
	FILE	*fp;
	int	x;
	long	l;

	fp = fopen ("download.req", "r+");
	if (fp == NULL) return;		/* No download file. Ignore it */

	fread (download_list, sizeof (unsigned long), MAXDOWNLOAD, fp);
	if (ferror (fp)) {		/* Read Error. Ignore it */
		fclose (fp);
		return;
	}

	/* If this request was the active one, clear it */
	if (download_active == file_id) download_active = 0xFFFFFFFF;

	/* Locate the requested ID in the download_list */
	for (x = 0; x < MAXDOWNLOAD; x++) {
		if (download_list [x] == file_id) break;
	}

	if (download_list [x] == file_id) {
		/* Clear the entry in the list and the file */
		download_list [x] = 0xFFFFFFFF;
		l = sizeof (unsigned long) * x;
		fseek (fp, l, SEEK_SET);
		fwrite (&download_list [x], sizeof (unsigned long), 1, fp);
	}
	fclose (fp);
}

/*
 *	Add a file to the download_list and file
 */
void AddDownloadFile (unsigned long fileid)
{
	FILE	*fp;
	int	x;
	long	l;

	fp = fopen ("download.req", "r+");
	if (fp == NULL) return;		/* No file. Ignore it */

	fread (download_list, sizeof (unsigned long), MAXDOWNLOAD, fp);
	if (ferror (fp)) {		/* Read error.  Ignore it */
		fclose (fp);
		return;
	}

	/* Locate a empty slot in the table */
	for (x = 0; x < MAXDOWNLOAD && download_list [x] != 0xFFFFFFFF; x++) ;
	if (download_list [x] == 0xFFFFFFFF) {
		/* An empty slot. Use it and update the file */
		download_list [x] = fileid;
		l = sizeof (unsigned long) * x;
		fseek (fp, l, SEEK_SET);
		fwrite (&download_list [x], sizeof (unsigned long), 1, fp);
	}
	fclose (fp);
}


/*
 *	Creates and sends a request to the satellite
 */
void UpdateRequest (unsigned int REQUEST_TYPE)
{
	static int		req_len;

	/* Create a new request */
	req_len = 0;

	/* Check for files first */
	/* Check if a file is in the download database */
	CheckDownloadFile ();

	if (download_active != 0xFFFFFFFF) {
		id = download_active;	/* Get this file */
		FillFile (&req_len);

		/* Check if this file should be requested */
		if (req_len == 0) RemoveDownloadFile (id);
	}
 	
	if (REQUEST_TYPE == REQUEST_ANY && req_len == 0) {
            /* a directory update request may go */
	    if (directoryOld || of [MAXFILES].holes > 1) {
	        /* The directory needs an update. Create request */
		FillDirectory ();
		if (req_len != 0) id = -2;	/* Requesting Directory */ }
	    else 
	        id = -1 ; /* Nothing is being requested */
	}
	/* Return if the request buffer is empty */
	if (req_len == 0) {
		return;
	}
}

/*
 *	Update status counters in the status line on the display
 */
void UpdateStatus(void)
{
	static int lastTotal         = -1;
	static int lastDir           = -1;
	static int lastFile          = -1;
	static int lastTLM           = -1;
	static int lastCRC           = -1;
	static int lastDHoles        = -1;
	static int lastdirectoryOld  = -1;
	static int lastrequest_no    = -1;
	char label[30];

	if (of [MAXFILES].holes != lastDHoles ||
	    directoryOld != lastdirectoryOld)
	{
		lastDHoles = of [MAXFILES].holes;
		if (directoryOld) {
			sprintf (label, "DIR: Old    ");
		} else if (lastDHoles == 1)
			sprintf (label, "DIR: Update ");
		else
			sprintf (label, "DIR: %d Parts ", lastDHoles - 1);
		writestatus (0, label);
	}

	if (lastrequest_no != id) {
		lastrequest_no = id;
		switch (id) {
			case -1: writestatus (15, "req:------ "); break;
			case -2: writestatus (15, "req:DIR    "); break;
			default:
				sprintf (label, "req:%-5lx ", id);
				writestatus (15, label);
		}
	}

	if (bytes != lastTotal)
	{
		sprintf(label, "T:%06d", bytes);
		writestatus (40, label);
		lastTotal = bytes;
	}

	if (fileBytes != lastFile)
	{
		sprintf(label, "F:%06d", fileBytes);
		writestatus (49, label);
		lastFile = fileBytes;
	}

	if (dirBytes != lastDir)
	{
		sprintf(label, "D:%06d", dirBytes);
		writestatus (58, label);
		lastDir = dirBytes;
	}

	if (tlmBytes != lastTLM)
	{
		sprintf(label, "T:%05d", tlmBytes);
		writestatus (67, label);
		lastTLM = tlmBytes;
	}

	if (crcErrors != lastCRC)
	{
		sprintf(label, "E:%03d", crcErrors);
		writestatus (75, label);
		lastCRC = crcErrors;
	}
}

/*
 * Hole list maintenance
 */
int LoadHoleList(int f, char *holeName, char *fileName)
{
	FILE *holeFile;
	FILE *fileFile;
	HOLE *hole;
	HOLE *prevHole;
	int i;

	holeFile = fopen(holeName, "r");
	fileFile = fopen(fileName, "r");

	/*
	 * handle the case where a .act file or a .hol file exist without
	 * the other. Usually caused by a crash.
	 */

	if (holeFile != NULL && fileFile != NULL)
	{
		fscanf(holeFile, "%d %*s %*s %*s", &of[f].hdrSeen);
		fscanf(holeFile, "%d %*s %*s %*s", &of[f].fileSize);
		fscanf(holeFile, "%d %*s", &of[f].holes);

		prevHole = NULL;

		for (i = 0; i < of[f].holes; i++)
		{
			hole           = (HOLE *)malloc(sizeof(HOLE));
			hole->nextHole = NULL;
			fscanf(holeFile, "%u, %u", &hole->start, &hole->finish);

			if (prevHole == NULL)
				of[f].firstHole = hole;
			else
				prevHole->nextHole = hole;

			prevHole = hole;
		}
	
		fclose(holeFile);
		fclose(fileFile);

		return(1);
	}
	else
	{
		if (fileFile != NULL) fclose(fileFile);
		if (holeFile != NULL) fclose(holeFile);
	
		/* create an empty hole list */
		of[f].firstHole           = (HOLE *)malloc(sizeof(HOLE));
		of[f].firstHole->start    = 0;
		of[f].firstHole->finish   = 0xFFFFFFFF;
		of[f].firstHole->nextHole = NULL;
		of[f].holes               = 1;

		return(0);
	}
}

void LoadDirectoryHoleList(void)
{
	FILE *holeFile;
	HOLE *hole;
	HOLE *prevHole;
	time_t t;
	int holes;
	unsigned oldest;
	unsigned start, finish;

	/* prune back the list */
	time(&t);
	oldest = (unsigned)t - (60 * 60 * 24 * maxDays);

	if ((holeFile = fopen("pfhdir.hol", "r")) != NULL)
	{
		fscanf(holeFile, "%d %*s %*s %*s", &of[MAXFILES].hdrSeen);
		fscanf(holeFile, "%d %*s %*s %*s", &of[MAXFILES].fileSize);
		fscanf(holeFile, "%d %*s", &holes);

		prevHole = NULL;

		of[MAXFILES].holes = 0;

		while (holes-- > 0)
		{
			fscanf(holeFile, "%u, %u", &start, &finish);

			if (finish > oldest)
			{
				hole           = (HOLE *)malloc(sizeof(HOLE));
				hole->nextHole = NULL;

				hole->start    = (start < oldest) ? oldest : start;
				hole->finish   = finish;

				if (prevHole == NULL)
					of[MAXFILES].firstHole = hole;
				else
					prevHole->nextHole = hole;

				prevHole = hole;

				of[MAXFILES].holes++;
			}
		}
	
		fclose(holeFile);
	}
	else
	{
		/* create an empty hole list */
		of[MAXFILES].firstHole           = (HOLE *)malloc(sizeof(HOLE));
		of[MAXFILES].firstHole->start    = oldest;
		of[MAXFILES].firstHole->finish   = 0xFFFFFFFF;
		of[MAXFILES].firstHole->nextHole = NULL;
		of[MAXFILES].holes               = 1;

		/* use file size as file number - start at 1 */
		of[MAXFILES].fileSize            = 1;
	}
}

void SaveHoleList(int f, char *fileName)
{
	char buffer[40];
	FILE *holeFile;
	HOLE *hole;

	if ((holeFile = fopen(fileName, "w")) == NULL)
	{
		sprintf(buffer, "Cannot open file %s", fileName);
		MessageBox(buffer);
		return;
	}
	
	fprintf(holeFile, "%d pfh header received\n", of[f].hdrSeen);
	fprintf(holeFile, "%d pfh file length\n", of[f].fileSize);
	fprintf(holeFile, "%d holes\n", of[f].holes);

	hole = of[f].firstHole;

	while (hole != NULL)
	{
		fprintf(holeFile, "%u, %u\n", hole->start, hole->finish);
		hole = hole->nextHole;
	}

	fclose(holeFile);
}

int UpdateHoleList(int f, unsigned start, unsigned finish)
{
	HOLE *h, *p, *n;

	for (p = NULL, h = of[f].firstHole; h != NULL; p = h, h = h->nextHole)
	{
		/* see if this is a candidate */
		if (start <= h->finish && finish >= h->start)
		{
			if (start == h->start && finish == h->finish)
			{
				/* exact match - will remove all of it */
				if (p == NULL)
					of[f].firstHole = h->nextHole;
				else
					p->nextHole = h->nextHole;

				of[f].holes--;
				free((char *)h);
				return(FILL_UNIQUE);
			}

			if (start <= h->start && finish >= h->finish)
			{
				/* larger - will remove all of it */
				if (p == NULL)
					of[f].firstHole = h->nextHole;
				else
					p->nextHole = h->nextHole;

				of[f].holes--;
				free((char *)h);
				return(FILL_OVERLAP);
			}

			if (start == h->start && finish < h->finish)
			{
				/* remove the front part */
				h->start = finish + 1;
				return(FILL_UNIQUE);
			}

			if (start < h->start && finish < h->finish)
			{
				/* remove the front part */
				h->start = finish + 1;
				return(FILL_OVERLAP);
			}

			if (start > h->start && finish == h->finish)
			{
				/* will remove the end */
				h->finish = start - 1;
				return(FILL_UNIQUE);
			}

			if (start > h->start && finish > h->finish)
			{
				/* will remove the end */
				h->finish = start - 1;
				return(FILL_OVERLAP);
			}

			if (start > h->start && finish < h->finish)
			{
				/* remove the middle */
				n = (HOLE *)malloc(sizeof(HOLE));
				n->start    = finish + 1;
				n->finish   = h->finish;
				n->nextHole = h->nextHole;

				/* change the current hole for the first part */
				h->finish   = start - 1;
				h->nextHole = n;

				/* one more hole */
				of[f].holes++;
				return(FILL_UNIQUE);
			}
		}
	}

	return(FILL_NONE);
}

/*
 * Message file maintenance
 */
void LoadFile(int fileId, int f)
{
	char buffer[80];
	char holeName[40];
	char fileName[40];

	/* save this file id */
	of[f].fileId = fileId;

	/* load the hole file and open the data file */
	sprintf(fileName, "%x.act", of[f].fileId);
	sprintf(holeName, "%x.hol", of[f].fileId);

	if (LoadHoleList(f, holeName, fileName))
	{
		if ((of[f].file = fopen(fileName, "r+")) == NULL)
		{
			sprintf(buffer, "Cannot open file %s", fileName);
			MessageBox(buffer);
		}
	}
	else
	{
		if ((of[f].file = fopen(fileName, "w+")) == NULL)
		{
			sprintf(buffer, "Cannot open file %s", fileName);
			MessageBox(buffer);
		}
	}
}

void SaveFile(int f)
{
	HOLE *hole;
	HOLE *next;
	char fileName[40];

	if (of[f].file != NULL)
	{
		sprintf (writebuf, "saving file %x", of[f].fileId);
		writelog (2, writebuf);

		/* close the data file */
		fclose(of[f].file);

		/* write out the hole list */
		sprintf(fileName, "%x.hol", of[f].fileId);
		SaveHoleList(f, fileName);

		/* clean up the file table */
		of[f].fileId   = 0;
		of[f].file     = NULL;
		of[f].hdrSeen  = 0;
		of[f].fileSize = 0;
		of[f].holes    = 0;

		/* free the hole list */
		hole = of[f].firstHole;

		while (hole != NULL)
		{
			next = hole->nextHole;
			free((char *)hole);
			hole = next;
		}

		of[f].firstHole = NULL;
	}
}

void CheckDownloaded(int f)
{
	int headerSize;
	HEADER *hdr;
	HOLE *h;
	unsigned char *buf;
	char oldName[80];
	char newName[80];
	char holeName[80];

	/* see if we have the header */
	if (of[f].hdrSeen == 0)
	{
		h = of[f].firstHole;

		if (h->start != 0)
		{
			/* read the header into a buffer */
			if ((buf = malloc(h->start)) != NULL)
			{
				fseek(of[f].file, 0L, SEEK_SET);
				fread(buf, 1, h->start, of[f].file);

				if ((hdr = ExtractHeader(buf, h->start, &headerSize)) != NULL)
				{
					of[f].hdrSeen  = 1;
					of[f].fileSize = hdr->fileSize;
					UpdateHoleList(f, (unsigned)of[f].fileSize, (unsigned)0xFFFFFFFF);
					free((char *)hdr);
				}

				free(buf);
			}
		}
	}

	/* see if we now have the complete file */
	if (of[f].holes == 0 && of[f].hdrSeen == 1)
	{
		unsigned long fileId;

		/* close the current file */
		fileId = of[f].fileId;
		SaveFile(f);

		/* rename the file */
		sprintf(oldName, "%lx.act", fileId);
		sprintf(newName, "%lx.dl", fileId);
		rename(oldName, newName);

		/* remove the hole file */
		sprintf(holeName, "%lx.hol", fileId);
		unlink(holeName);

		/* Remove the file from the download_list */
		RemoveDownloadFile (fileId);

		/* let the user know */
		sprintf(writebuf, "%lx downloaded", fileId);
		writelog (2, writebuf);

		/* Check if this file completes the active request */
		/* send the satellite a new request */
		/* This is a optimizing hack. ???? */
/* This hack is disabled. ?????
		if (id == fileId) {
			in_queue = 0;		/ * Clear the queue flag * /
			UpdateRequest ();	/ * Create a new request * /
		}
*/
	}
}

void BroadcastFile(unsigned char *buffer, int length)
{
	static unsigned HeardId = 0;
	static unsigned DupId   = 0;
	FILEHEADER *fh;
	unsigned char *data;
	unsigned dataOffset;
	unsigned dataLength;
	char fileName[80];
	char txt[100];
	FILE *fp;
	int i;

	/* crc validation */
	if (!CheckCRC(buffer, length))
	{
		crcErrors++;
		UpdateStatus();
		return;
	}

	/* point to the header */
	fh = (FILEHEADER *)buffer;

	/* point to the data */
	data       = buffer + sizeof(FILEHEADER);
	dataOffset = fh->wOffset + (fh->nOffsetHigh << 16);
	dataLength = length - sizeof(FILEHEADER) - CRCLENGTH;
	
	/* inform the user */
	if (fh->fileId != HeardId)
	{
		sprintf  (writebuf, "heard message %x", fh->fileId);
		writelog (2, writebuf);
		HeardId = fh->fileId;
		DupId   = -1;
	}

	/* see if the file already downloaded */
	sprintf(fileName, "%x.dl", fh->fileId);

	if ((fp = fopen(fileName, "r")) != NULL)
	{
		fclose(fp);

		if (fh->fileId != DupId)
		{
			sprintf  (writebuf, "%x is already downloaded", fh->fileId);
			writelog (2, writebuf);
			DupId = fh->fileId;
		}

		return;
	}

	/* see if the file is in the current list of files */
	for (i = 0; i < MAXFILES; i++)
		if (of[i].fileId == fh->fileId)
			break;

	/* if it is not there then we must load it */
	if (i == MAXFILES)
	{
		/* Locate a empty slot before closing an active slot ???? */

		/* save current file if used */
		SaveFile(nextof);

		/* now use this slot */
		LoadFile(fh->fileId, nextof);
		
		/* setup i for this entry */
		i = nextof;

		/* make sure we round robin the open files */
		nextof = (nextof + 1) % MAXFILES;
	}

	if (UpdateHoleList(i, (unsigned)dataOffset, (unsigned)(dataOffset + dataLength - 1)) != FILL_NONE)
	{
		/* write the data */
		fseek(of[i].file, dataOffset, SEEK_SET);
		fwrite(data, 1, dataLength, of[i].file);

		/* Update the download window */
		sprintf (txt, "%4x  %4d  %6d  %s",
		              of[i].fileId, of[i].holes, dataOffset,
		              (of[i].hdrSeen == 1 ? "Y" : "N"));
		writedownload (i + 1, txt);

		/* see if it has all been downloaded */
		CheckDownloaded(i);

		/* Check if the file is completed */
		if (of[i].file == NULL) {
			writedownload (i + 1, "");
		}
	}
}

/*
 *	Directory Maintenance
 */
void WriteDirectory(unsigned char *data, int dataLength)
{
	char buffer[80];
	char fileName[40];
	struct stat statbuf;
	FILE *fp;

	sprintf(fileName, "pb__%04d.pfh", of[MAXFILES].fileSize);

	/* write the data to the end of the current file */
	if ((fp = fopen(fileName, "a")) == NULL)
	{
		sprintf(buffer, "Cannot open file %s", fileName);
		MessageBox(buffer);
		return;
	}

	fwrite(data, 1, dataLength, fp);
	fclose(fp);

	stat(fileName, &statbuf);

	/* limit file size to 20000 bytes */
	if (statbuf.st_size > 20000) of[MAXFILES].fileSize++;
}

void BroadcastDirectory(unsigned char *buffer, int length)
{
	static unsigned int  fragmentId     = 0;
	static unsigned int  fragmentOffset = 0;
	static unsigned char fragmentBuffer[1024];
	DIRHEADER	*dh;
	int		headerSize;
	HEADER		*hdr;
	LOADLIST	*ll;
	unsigned char 	*data;
	int		dataLength;
	int		i;

	/* crc validation */
	if (!CheckCRC(buffer, length))
	{
		crcErrors++;
		UpdateStatus();
		return;
	}

	/* point to the header */
	dh = (DIRHEADER *)buffer;

	/* check for fragmented directory header */
	if (dh->offset != 0)
	{
		/* see if this is the next fragment */
		if (fragmentId == dh->fileId && dh->offset == fragmentOffset)
		{
			/* append this fragment */
			data       = buffer + sizeof(DIRHEADER);
			dataLength = length - sizeof(DIRHEADER) - CRCLENGTH;

			for (i = 0; i < dataLength; i++)
				fragmentBuffer[fragmentOffset + i] = data[i];

			fragmentOffset += dataLength;

			if ((dh->flags & LASTBYTEFLAG) != LASTBYTEFLAG)
			{
				data       = fragmentBuffer;
				dataLength = fragmentOffset;
			}
			else
			{
				return;
			}
		}
		else
		{
			fragmentId     = 0;
			fragmentOffset = 0;
			return;
		}
	}
	else if ((dh->flags & LASTBYTEFLAG) != LASTBYTEFLAG)
	{
		/* copy the fragment */
		data       = buffer + sizeof(DIRHEADER);
		dataLength = length - sizeof(DIRHEADER) - CRCLENGTH;

		for (i = 0; i < dataLength; i++)
			fragmentBuffer[i] = data[i];

		/* start a new fragment */
		fragmentId     = dh->fileId;
		fragmentOffset = dataLength;

		return;
	}
	else
	{
		/* its all there -  point to the data */
		data       = buffer + sizeof(DIRHEADER);
		dataLength = length - sizeof(DIRHEADER) - CRCLENGTH;
	}

	if ((dh->flags & LASTBYTEFLAG) == LASTBYTEFLAG)
	{
		/* reset fragmentation */
		fragmentId     = 0;
		fragmentOffset = 0;

		/* try to extract the header */
		if ((hdr = ExtractHeader(data, dataLength, &headerSize)) != NULL)
		{
			/* inform the user */
			if (strlen(hdr->title) == 0)
				sprintf(writebuf, "dir: %lx: to:%s  %s",
					hdr->fileId,
					hdr->destination,
					hdr->fileName);
			else
				sprintf(writebuf, "dir: %lx: to:%s  %s",
					hdr->fileId,
					hdr->destination,
					hdr->title);

			notprintable (writebuf);
			writelog (2, writebuf);

			/* update the directory  if needed */
			if (UpdateHoleList(MAXFILES, dh->tOld, dh->tNew) == FILL_UNIQUE) {
				WriteDirectory(data, dataLength);

				/* Check if the directory is complete */
				/* If so end a DIR request */
/*  This is a optimizing hack.  Currently disabled ??????
				if (of [MAXFILES].holes == 1 && request_no == REQUEST_DIR)
					UpdateRequest (REQUEST_NONE);
*/

				/* Check if the just received file is to be downloaded */
				ll = autoload_list;
				for (;;ll = ll->next) {
					if (ll == NULL) break;

					if (ll->dest[0] != '\0' && ! strstr (hdr->destination, ll->dest)) continue;
					if (ll->src [0] != '\0' && ! strstr (hdr->source,      ll->src))  continue;
					if (! strstr (hdr->title,       ll->title)) continue;
					if (! strstr (hdr->keyWords,    ll->key))   continue;
					if (hdr->fileSize >             ll->size)   continue;

					/* Add this file to download_list */
					AddDownloadFile (hdr->fileId);
					writelog (2, "downloading ^^^^^^ file");
					break;
				}
			}
			free ((char *)hdr);

			/* Mark the directory as in KNOWN state now */
			directoryOld = 0;
		}
		else
		{
			sprintf  (writebuf, "** Bad directory header for %x", dh->fileId);
			writelog (1, writebuf);
		}
	}
}

/*
 *	decode a received frame.
 */
void ProcessFrame(void)
{
	int n=1, i;
	int via;
	unsigned char protocol;
	char toCall[10];
	char fromCall[10];
	char viaCall[10];
	char *s;

	/* Check for runts/pygmygs */
	if (bufSize < 15) {
		char	st[10];
		int	x;
		sprintf (writebuf, "Runt:");
		for (x = 0; x < bufSize; x++) {
			sprintf (st, " %02x", buf [x]);
			strcat (writebuf, st);
		}
		writelog (1, writebuf);
		return;
	}
	
	bytes += bufSize - 1 ;

	/* decode the to/from address */
	/* dont expect via address, but saves last if any */
	via = ConvertCall(buf + n, toCall);
	n += 7;

	via = ConvertCall(buf + n, fromCall);
	n += 7;

	while (via)
	{
		via = ConvertCall(buf + n, viaCall);
		n += 7;
	}

	/* Check frame trace */
	if (traceFlag > 0) {
		sprintf (writebuf, ">>> %s -=> %s  [%02x] [%02x]",
		fromCall, toCall, buf [n], buf [n+1]);
		writelog (1, writebuf);
		if (traceFlag > 1) {
			int	x;
			char	st[10];
			for (x = n + 2; x < bufSize; x++) {
				sprintf (st, " %02x", buf [x]);
				strcat (writebuf, st);
			}
			writelog (1, writebuf);
		}
	}

	if (fpkiss != NULL)
	{
		fputc(0300, fpkiss);		/* FEND */
		fputc(0x00, fpkiss);		/* KISS Data */

		for (i = 1; i < bufSize; i++)
		{
			switch (buf[i])
			{
				case 0300:	/* FEND */
					fputc(0333, fpkiss);	/* FESC */
					fputc(0334, fpkiss);	/* FESC_END */
					break;
				case 0333:	/* FESC */
					fputc(0333, fpkiss);	/* FESC */
					fputc(0335, fpkiss);	/* FESC_ESC */
					break;
				default:
					fputc(buf[i], fpkiss);
					break;
			}
		}

		fputc(0300, fpkiss);		/* FEND */
	}


	/* check for a UI frame */
	if ((buf[n] & 0xEF) != 0003)
	{
		/* Ignore non UI frames */
		UpdateStatus ();
		return;
	}

	n++;
	protocol = buf[n++];

	/* see if the frame is a broadcast frame */	
	if (strcmp(toCall, "QST-1") == 0)
	{
		switch (protocol)
		{
		case PID_FILE:
			fileBytes += bufSize - n;
			BroadcastFile(buf + n,  bufSize - n);
			break;
		case PID_DIRECTORY:
			dirBytes += bufSize - n;
			BroadcastDirectory(buf + n, bufSize - n);
			break;
		default:
			break;
		}
	}
	else if (strcmp(toCall, "PBLIST") == 0)
	{
		buf[bufSize] = '\0';
		sprintf  (writebuf, "%s", buf + n);
		notprintable (writebuf);
		writelog (1, writebuf);

		/* Check if a request has not been received by the satellite */
		s = strstr (buf + n, MyCall);
		if (s != NULL)  /* I might be in the queue */ {
			/* Check that the char after MyCall is a SP, a '\' or a CR */
			s += strlen (MyCall);
			if (*s == ' ' || *s == '\\' ||*s == '\r')   {
                           /* I'm in the queue */
                           if (*s == '\\') {
                              /* I have a DIR request at the satellite 
                                 see if a FILE request should go instead */
                              UpdateRequest (REQUEST_FILE) ; }
                           else /* it's a FILE request - do nothing */
                              ; }
                        else /* it wasn't my call after all */
                           UpdateRequest (REQUEST_ANY) ; }
                else /* not in the queue */
                   UpdateRequest (REQUEST_ANY) ;

	}
	else if (strcmp(toCall, "TLM") == 0)
	{
		tlmBytes += bufSize - n;

		if (!CheckCRC(buf + n, bufSize - n))
		{
			crcErrors++;
			UpdateStatus();
			return;
		}
	}
	else if (strcmp (toCall, "BSTAT") == 0)
	{
		buf [bufSize - 1] = '\0';
		sprintf (writebuf, "%s", buf + n);
		notprintable (writebuf);

		if (writebuf [0] == 'B' &&
		    writebuf [1] == ':' &&
		    writebuf [2] == ' ') {
			/* ByteCounter message */
			curByteRead  = dirBytes + fileBytes;
			curByteWrite = atol (writebuf + 3);

			if (lastByteRead == 0) {
				writelog (1, writebuf);
			} else {
				long l1, l2, l3 = 100L;
				l1 = (curByteRead  - lastByteRead);
				l2 = (curByteWrite - lastByteWrite);
				if (l2 != 0L) l3 = (100L * l1) / l2;
				sprintf (writebuf, "%s    (%ld / %ld = %ld%%)",
				         buf + n,
				         (curByteRead - lastByteRead),
				         (curByteWrite - lastByteWrite),
				         l3);
				notprintable (writebuf);
				writelog (1, writebuf);
			}
			lastByteRead  = curByteRead;
			lastByteWrite = curByteWrite;
		} else {
			writelog (1, writebuf);
		}
	}
	else if (strcmp (toCall, MyCall) == 0)
	{
		/* This packet is sent directly to ME */
		buf[bufSize] = '\0';
		sprintf  (writebuf, "%s", buf + n);
		notprintable (writebuf);
		writelog (1, writebuf);


		/* Check for negative responce for a FILE FILL request */
		if (protocol == PID_FILE &&
		    strncmp(buf + n, "NO -2 ", 6) == 0)
		{
			sprintf  (writebuf, "** File %lx not present, request cancelled", id);
			writelog (2, writebuf);
			RemoveDownloadFile (id);

			/* Make an new request. This is a optimization of */
			/* the protocol. I might wait for the PB: xxx list ???? */
			UpdateRequest (REQUEST_FILE);
		}

		/* Check for confirmation of a DIR FILL request */
		if (protocol == PID_DIRECTORY &&
		    strncmp (buf + n, "OK ", 3) == 0)
		{
			/* This information is currently not used ???? */
		}
	}
	else
	{
		buf [bufSize] = '\0';
		sprintf (writebuf, "%s", buf + n);
		notprintable (writebuf);
		writelog (1, writebuf);
	}

	UpdateStatus();
}

/*
 *	Input function when data is received on the input stream
 */
void GetFrame(void)
{
	if ((bufSize = recv (s_input, buf, MAXBUFFER, 0)) == -1) {
		perror ("recv") ;
		return ; }

		/* Check for read errors */
		if (bufSize == -1) stopFlag = 1;	/* Stop program */

		/* Process the received frame */
		time(&end_time);
		if (start_time == 0) start_time = end_time;
		ProcessFrame();
}

/*
 * Timeout has occured, flush the directory hole list.
 */
void TimeoutCb(void)
{
#ifdef FLUSH_FILES
	char fileName[40];
	int i;

	for (i = 0; i < MAXFILES; i++)
	{
		if (of[i].file != NULL)
		{
			/* write out the hole list */
			sprintf(fileName, "%x.hol", of[i].fileId);
			SaveHoleList(i, fileName);
		}
	}
#endif

	/* save the directory hole list */
	if (of[MAXFILES].firstHole != NULL)
		SaveHoleList(MAXFILES, "pfhdir.hol");
}

/*
 *	callback function when a another app calls us
 */
void CommandRequest(void)
{
	struct request request;

	if (recv(s_request, (char *)&request, sizeof(struct request), 0) == sizeof(struct request))
	{
		switch (request.type)
		{
			case REQ_TYPE_CANCEL:
				CancelFill();
				break;
			case REQ_TYPE_DIR:
/*				FillDirectory();	*/
				break;
			case REQ_TYPE_FILE:
/*				sprintf(id, "%lx", request.fileId);   */
/*				id = request.fileId;	*/
/*				FillFile(0);	*/
				break;
			case REQ_TYPE_REFILE:
				writelog (2, "download.req updated");
				/* If requester is idle check it */
/* ???? This hack is disabled.The request will be sent on next PB frame
				if (request_no == REQUEST_NONE)
					UpdateRequest (REQUEST_NONE);
*/
				break;
			case REQ_TYPE_QUIT:
				writelog (2, "Stop request received");
				stopFlag = 1;
				break;
		}
	}
	else
	{
		perror("recv - FillRequest");
	}
}

/*
 *	the user wants to exit this program
 */
void QuitCb(void)
{
	FILE *fp;
	struct tm *tm;
	int ave_rate = 0, i;


	/* save any open files */
	for (i = 0; i < MAXFILES; i++)
		SaveFile(i);

	/* save the directory hole list */
	if (of[MAXFILES].firstHole != NULL)
		SaveHoleList(MAXFILES, "pfhdir.hol");

	/* close the sockets */
	close(s_input);
	close(s_file);
	close(s_directory);

	/* write the statistics out */
	if (start_time != 0)
	{
		if ((fp = fopen("pb.log", "a")) != NULL)
		{
			if (end_time - start_time > 0)
				ave_rate = bytes / (end_time - start_time);

			tm = gmtime(&start_time);
			fprintf(fp, "%02d/%02d/%02d %02d:%02d:%02d - ",
				tm->tm_mday, tm->tm_mon + 1, tm->tm_year,
				tm->tm_hour, tm->tm_min, tm->tm_sec);

			tm = gmtime(&end_time);
			fprintf(fp, "%02d:%02d:%02d - total %d files %d dir %d tlm %d ave %d crc %d\n",
				tm->tm_hour, tm->tm_min, tm->tm_sec,
				bytes, fileBytes, dirBytes, tlmBytes, ave_rate, crcErrors);
	
			fclose(fp);
		}
	}
}


/*
 *	Construct a fill request for the directory.
 */
void FillDirectory(void)
{
	DIRREQUESTHEADER *request;
	DIRHOLEPAIR *hole;
	HOLE *h;
	unsigned char *buffer;
	int i;

        /* construct a hole list request */
	buffer = (unsigned char *)malloc (256);
        request = (DIRREQUESTHEADER *)buffer;
	request->flags     = 0x00 | VERSION | 0x10;
	request->blockSize = 244;

        hole = (DIRHOLEPAIR *)&buffer[sizeof(DIRREQUESTHEADER)];
	h = of[MAXFILES].firstHole;
	for (i = 0; i < 10 && i < of[MAXFILES].holes; i++)
	{
		hole->startTime = h->start;
		hole->endTime   = h->finish;
		h = h->nextHole;
		hole++;
	}

	if (sendto(s_directory, buffer,
		   sizeof(DIRREQUESTHEADER) + (i * sizeof(DIRHOLEPAIR)),
		   0, (struct sockaddr *)&dest, sizeof(dest)) == -1)
		perror ("send - fill dir") ;
	sprintf  (writebuf, "Fill Directory: fill %d holes.", i);
	writelog (2, writebuf);
	free (buffer);
}


/*
 *	construct a fill request for a file
 */
void FillFile(int *reqlen)
{
	unsigned int fileId;
	char fileName[80];
	unsigned char * buffer;
	REQUESTHEADER *request;
	HOLEPAIR *hole;
	HOLE *h;
	FILE *fp;
	int f, i;

	fileId = id;
	if (fileId == 0) return;

	/* see if the file already downloaded */
	sprintf(fileName, "%x.dl", fileId);
	if ((fp = fopen(fileName, "r")) != NULL)
	{
		fclose(fp);
		sprintf  (writebuf, "%x is already downloaded", fileId);
		writelog (2, writebuf);
		return;
	}

	/* see if the file is in the current list of files */
	for (f = 0; f < MAXFILES; f++)
		if (of[f].fileId == fileId)
			break;

	if (f == MAXFILES)
	{
		/* not there - must save if this is in use */
		SaveFile(nextof);

		/* now use this slot */
		LoadFile(fileId, nextof);

		/* make sure we round robin the files */
		f = nextof;
		nextof = (nextof + 1) % MAXFILES;
	}
	buffer = (unsigned char *)malloc(256);
	request = (REQUESTHEADER *)buffer ;
	if (of[f].holes == 1 && of[f].firstHole->start  == 0 &&
				of[f].firstHole->finish == 0xFFFFFFFF)
	{
		request->flags     = 0x00 | VERSION | 0x10;
		request->fileId    = fileId;
		request->blockSize = 244;
	
		if (sendto(s_file, buffer, sizeof(REQUESTHEADER), 0,
	                   (struct sockaddr *)&dest, sizeof(dest)) == -1)
                        perror("send - req. file");
                /* Update the *reqlen with the request header size */
                *reqlen += sizeof (REQUESTHEADER);


		sprintf  (writebuf, "Fill File: %x send file.", fileId);
		writelog (2, writebuf);
	}
	else
	{
		request->flags     = 0x02 | VERSION | 0x10;
		request->fileId    = fileId;
		request->blockSize = 244;

		hole = (HOLEPAIR *)&buffer[*reqlen + sizeof(REQUESTHEADER)];
		h = of[f].firstHole;

		for (i = 0; i < 10 && i < of[f].holes; i++)
		{
			hole->offset     = h->start;
			hole->offset_msb = h->start >> 16;
			hole->length     = 1 + (h->finish - h->start);
			h = h->nextHole;
			hole++;
		}
                if (sendto(s_file, buffer,
                            sizeof(REQUESTHEADER) + (i * sizeof(HOLEPAIR)), 0,
                            (struct sockaddr *)&dest, sizeof(dest)) == -1)
                        perror("send - fill file");
                /* Update the *reqlen with the request size */
                *reqlen +=  sizeof (REQUESTHEADER) +
                           (sizeof (HOLEPAIR) * i);


		sprintf  (writebuf, "Fill File: %x fill %d holes.", fileId, i);
		writelog (2, writebuf);
	}
	free (buffer);

	/* Update the status line */
	sprintf  (writebuf, "File: %-5x  ", fileId);
	writestatus (15, writebuf);
}



/*
 *	Cancel a Fill request		????
 */
void CancelFill(void)
{
#ifdef 0
	switch (request_state)
	{
		case REQUEST_DIR:
			writelog (2, "Cancel Fill: directory");
			break;
		case REQUEST_FILE:
			sprintf  (writebuf, "Cancel Fill: file %lx", id);
			writelog (2, writebuf);
			break;
		default:
			break;
	}

	request_state = REQUEST_NONE;
#endif
}

int main(int argc, char **argv)
{
	int 		n;

	fd_set		readfds;
	struct timeval	timeout;
	time_t		timeoutflush;

	FILE		*paramf;
	char		s1[100], s2[100];
	FILE		*autof;


	/* Read the parameter file pb.conf */
	if ((paramf = fopen ("pb.conf", "r")) == NULL) {
		perror ("pb.conf fopen");
		exit (1);
	}

	while (! feof (paramf)) {
		fscanf (paramf, "%100s %100s\n", s1, s2);

		if (! strcmp (s1, "SATELLITE"))   {strcpy (Satellite,   s2); continue;}
		if (! strcmp (s1, "SATNAME"))     {strcpy (SatName,     s2); continue;}
		if (! strcmp (s1, "MYCALL"))      {strcpy (MyCall,      s2); continue;}
		if (! strcmp (s1, "MAXDAYS"))     {strcpy (MaxDays,     s2); continue;}
		if (! strcmp (s1, "UDPPORT"))     {strcpy (UDPPort,     s2); continue;}
		if (! strcmp (s1, "KISSLOG"))     {strcpy (KissLog,     s2); continue;}

		printf ("Unknown config line :  <%s> = <%s>\n", s1, s2);
	}
	fclose (paramf);

	/* Process command line options */
	for (n = 1; n < argc; n++) {
		if (! strcmp (argv [n], "-s")) {strcpy (Satellite,   argv [++n]); continue;}
		if (! strcmp (argv [n], "-n")) {strcpy (SatName,     argv [++n]); continue;}
		if (! strcmp (argv [n], "-m")) {strcpy (MyCall,      argv [++n]); continue;}
		if (! strcmp (argv [n], "-M")) {strcpy (MaxDays,     argv [++n]); continue;}
		if (! strcmp (argv [n], "-U")) {strcpy (UDPPort,     argv [++n]); continue;}
		if (! strcmp (argv [n], "-k")) {strcpy (KissLog,     argv [++n]); continue;}
	}


	/*  Check if the needed parameters exist */
	if (Satellite [0] == '\0') {
		printf ("SATELLITE parameter is not defined\n");
		exit (1);
	}
	if (MyCall [0] == '\0') {
		printf ("MYCALL parameter is not defined\n");
		exit (1);
	}

	/* Convert string values to numbers */
	if (MaxDays [0] != '\0') {
		maxDays = atoi(MaxDays);
		if (maxDays <= 0) maxDays = 5;
	}
	if (UDPPort [0] != '\0') {
		UdpPort = atoi (UDPPort);
		if (UdpPort < 1024) UdpPort = 5100;
	}


	/* Create the packet addresses */
	strcat(Satellite, "-11");
	MakeAddress(MyCall, &src);
	MakeAddress(Satellite, &dest);

	/* Load the autoload.dat file into memory */
	if ((autof = fopen ("autoload.dat", "r")) != NULL) {
		LOADLIST *ll = NULL;
		for (;;) {
			if (feof (autof)) break;

			autoload_list = malloc (sizeof (LOADLIST));
			autoload_list->next = ll;
			ll = autoload_list;
				
			fscanf (autof, "%s %s %s %s %d\n", ll->dest, ll->src,
				ll->title, ll->key, &ll->size);

			if (ll->dest  [0] == '*') ll->dest  [0] = '\0';
			if (ll->src   [0] == '*') ll->src   [0] = '\0';
			if (ll->title [0] == '*') ll->title [0] = '\0';
			if (ll->key   [0] == '*') ll->key   [0] = '\0';
			if (ll->size      == 0)   ll->size      = 0xFFFFFFFF;
		}
		fclose (autof);
	}

	id = -1;
	/* initialize the open files structure */
	for (n = 0; n < MAXFILES; n++)
	{
		of[n].fileId    = 0;
		of[n].file      = NULL;
		of[n].hdrSeen   = 0;
		of[n].fileSize  = 0;
		of[n].holes     = 0;
		of[n].firstHole = NULL;
	}

        if (ax25_config_load_ports () == 0) {
                perror ("no AX.25 ports configured") ;
                return 1 ; }

        if ((s_input = socket (AF_INET, SOCK_PACKET, htons (ETH_P_AX25))) == -1) {
                perror ("socket") ;
                return 1 ; }

        /* open up the AX25 datagram socket to send file requests */
        if ((s_file = socket(AF_AX25, SOCK_RAW, PID_FILE)) == -1)
        {
                perror("socket");
                return(1);
        }

        if (bind(s_file, (struct sockaddr *)&src, sizeof(src)) == -1)
        {
                perror("bind");
                return(1);
        }

        /* open up the AX25 datagram socket to send directory requests */
        if ((s_directory = socket(AF_AX25, SOCK_RAW, PID_DIRECTORY)) == -1)
        {
                perror("socket");
                return(1);
        }

        if (bind(s_directory, (struct sockaddr *)&src, sizeof(src)) == -1)
        {
                perror("bind");
                return(1);
        }

	/* open up a udp socket to receive file fill requests from other apps */
	memset((char *)&request_addr, 0, sizeof(request_addr));
	request_addr.sin_family      = AF_INET;
	request_addr.sin_addr.s_addr = INADDR_ANY;
	request_addr.sin_port        = htons(UdpPort);

	if ((s_request = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
	{
		perror("socket - s_request");
	}
		
	if (bind(s_request, (struct sockaddr *)&request_addr, sizeof(request_addr)) == -1)
	{
		perror("bind - s_request");
	}

	/* Open the KissLog file */
	if (KissLog [0] != 0) {
		time_t timenow ;
		struct tm *tm ;
		time (&timenow) ;
		tm = gmtime (&timenow) ;
		sprintf (KissLog, "%02d%02d%02d%02d.kss", tm->tm_year, tm->tm_mon+1, tm->tm_mday, tm->tm_hour) ;
		if ((fpkiss = fopen(KissLog, "ab")) == NULL)  {
			perror ("Error opening the KissLog file");
		}
	}


	/* Load the Directory Hole List into memory */
	LoadDirectoryHoleList ();

	/* Load the download_list from the download.req file */
	CheckDownloadFile ();

	/* Initialicer curses skrmstyringen */
	initscr ();
	cbreak ();		/* Set the keyboard in raw mode */
	noecho ();

	scr_download = newwin (11, 24,  0,  0);
	scr_files    = newwin (11, 54,  0, 25);
	scr_log      = newwin (12, 80, 12,  0);
	scr_status   = newwin ( 1, 80, 24,  0);

	scrollok (scr_files, TRUE);
	scrollok (scr_log,   TRUE);

	/* Initialise the screen */
	/* Header in the download window */
	mvwprintw (scr_download, 0, 0, "File Holes  Offset  Head");
	wrefresh (scr_download);

	/* Create frames in the background */
	wattron   (stdscr, A_REVERSE);
	mvwprintw (stdscr, 11, 0, "%80s", SatName);
	for (n = 0; n < 11; n++) mvwprintw (stdscr, n, 24, " ");
	wattroff  (stdscr, A_REVERSE);
	wrefresh  (stdscr);

	/* Invert the status line */
	wattron   (scr_status, A_REVERSE);
	mvwprintw (scr_status, 0, 0, "%80s", "");
	wrefresh  (scr_status);

	UpdateStatus();

	/* Main loop. Start select on DataInput, updInput, Keyboard and timeout */
	timeout.tv_sec  = TIMEOUT;
	timeout.tv_usec = 0;
	timeoutflush    = TIMEOUTFLUSH + time (0);
	while (! stopFlag) {
		/* Prepare select */
		FD_ZERO (&readfds);
		FD_SET (s_input,        &readfds);
		FD_SET (s_request,      &readfds);
		FD_SET (fileno (stdin), &readfds);
		n = select (32, &readfds, NULL, NULL, &timeout);
		if (n < 0) {
			perror ("select");
			exit (1);
		}

		/* Check if the TimeoutCb has to be run */
		if (time (0) > timeoutflush) {
			TimeoutCb ();	/* Call timeout process */

			/* Setup next time */
			timeoutflush = TIMEOUTFLUSH + time (0);
		}

		/* Reset timeout values */
		timeout.tv_sec  = TIMEOUT;
		timeout.tv_usec = 0;

		/* Chech which descriptors came out with some data */
		if (FD_ISSET (s_input,        &readfds)) GetFrame ();
		if (FD_ISSET (s_request,      &readfds)) CommandRequest ();

		if (FD_ISSET (fileno (stdin), &readfds)) {
			/* Read the char from the keyboard */
			n = fgetc (stdin);

			switch (n) {
			    case 'Q':	/* Q: Leave the program */
			    case 'q':
				stopFlag = 1;
				break;
			    case 't':	/* t: Toggle trace option */
				if (traceFlag != 1)
					traceFlag = 1;
				else
					traceFlag = 0;
				break;
			    case 'T':	/* T: Toggle trace option */
				if (traceFlag != 2)
					traceFlag = 2;
				else
					traceFlag = 0;
				break;
			    case 'd':	/* D: Directory Fill */
			    case 'D':
/*				FillDirectory ();	/ * ???? */
				break;
			}
		}
	}

	/* Terminate the program */
	writelog (1, "");
	writelog (1, "");
	writelog (1, "Program terminated normally");
	writelog (2, "");
	writelog (2, "Program terminated normally");
	writestatus (0, "Program terminated          ");

	QuitCb ();		/* Flush buffers and close all files */
	endwin ();
	return(0);
}

